package com.maroontress.coverture;

import com.maroontress.coverture.gcda.FunctionDataRecord;
import com.maroontress.coverture.gcno.AnnounceFunctionRecord;
import com.maroontress.coverture.gcno.ArcRecord;
import com.maroontress.coverture.gcno.ArcsRecord;
import com.maroontress.coverture.gcno.FunctionGraphRecord;
import com.maroontress.coverture.gcno.LineRecord;
import com.maroontress.coverture.gcno.LinesRecord;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.LinkedList;

/**
   ؿդǤ
*/
public final class FunctionGraph {

    /** ̻ҤǤ */
    private int id;

    /** ؿΥåǤ */
    private int checksum;

    /** ؿ̾Ǥ */
    private String functionName;

    /** ɤΥե̾Ǥ */
    private String sourceFile;

    /** ؿи륽ɤιԿǤ */
    private int lineNumber;

    /** ؿܥ֥åǤ */
    private Block[] blocks;

    /** ٤ƤΥθĿǤ */
    private int totalArcCount;

    /**
       ΥθĿǤΥϡ㳰longjmp()ˤäơ
       ߤδؿȴϩбޤ
    */
    private int fakeArcCount;

    /** ¹ԲȽƤ륢ΥꥹȤǤ */
    private ArrayList<Arc> solvedArcs;

    /** ¹ԲʥΥꥹȤǤ */
    private ArrayList<Arc> unsolvedArcs;

    /** ƤӽФ줿Ǥ */
    private long calledCount;

    /** äǤ */
    private long returnedCount;

    /** ¹Ԥ줿֥åȽиˤθĿǤ */
    private int executedBlockCount;

    /** եդ褷Ƥ뤫ɤɽޤ */
    private boolean solved;

    /**
       ꥹȤˤδؿդɲä٤ƤΥ֥åιֹ
       μ¹Բޡޤ

       եդ褷Ƥʤϲ⤷ޤ

       @param sourceList ꥹ
    */
    public void addLineCounts(final SourceList sourceList) {
	if (!solved) {
	    return;
	}
	Source source = sourceList.getSource(sourceFile);
	source.addFunctionGraph(this);
	for (Block b : blocks) {
	    b.addLineCounts(sourceList);
	}
    }

    /**
       ؿդXMLǽϤޤ

       @param out 
    */
    public void printXML(final PrintWriter out) {
	int complexityWithFake = totalArcCount - blocks.length + 2;
	int complexity = complexityWithFake - fakeArcCount;
	out.printf("<functionGraph id='%d' checksum='0x%x' functionName='%s'"
		   + " sourceFile='%s' lineNumber='%d'"
		   + " complexity='%d' complexityWithFake='%d'",
		   id, checksum, XML.escape(functionName),
		   XML.escape(sourceFile), lineNumber,
		   complexity, complexityWithFake);
	if (solved) {
	    out.printf(" called='%d' returned='%d' executedBlocks='%d'",
		       calledCount, returnedCount, executedBlockCount);
	}
	out.printf(" allBlocks='%d'>\n", getBlockCount());
	for (Block b : blocks) {
	    b.printXML(out);
	}
	out.printf("</functionGraph>\n");
    }

    /**
       ؿե쥳ɤ饤󥹥󥹤ޤ

       @param rec ؿե쥳
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    public FunctionGraph(final FunctionGraphRecord rec)
	throws CorruptedFileException {
	AnnounceFunctionRecord announce = rec.getAnnounce();
	id = announce.getId();
	checksum = announce.getChecksum();
	functionName = announce.getFunctionName();
	sourceFile = announce.getSourceFile();
	lineNumber = announce.getLineNumber();
	solvedArcs = new ArrayList<Arc>();
	unsolvedArcs = new ArrayList<Arc>();

	int[] blockFlags = rec.getBlocks().getFlags();
	blocks = new Block[blockFlags.length];
	for (int k = 0; k < blockFlags.length; ++k) {
	    blocks[k] = new Block(k, blockFlags[k]);
	}

	ArcsRecord[] arcs = rec.getArcs();
	for (ArcsRecord e : arcs) {
	    addArcsRecord(e);
	}

	LinesRecord[] lines = rec.getLines();
	for (LinesRecord e : lines) {
	    addLinesRecord(e);
	}

	if (blocks.length < 2) {
	    throw new CorruptedFileException("lacks entry and/or exit blocks");
	}
	Block entryBlock = blocks[0];
	Block exitBlock = blocks[blocks.length - 1];
	if (entryBlock.getInArcs().size() != 0) {
	    throw new CorruptedFileException("has arcs to entry block");
	}
	if (exitBlock.getOutArcs().size() != 0) {
	    throw new CorruptedFileException("has arcs from exit block");
	}
	for (Block e : blocks) {
	    e.presolve();
	}
    }

    /**
       ARCS쥳ɤ饢ơؿդɲäޤ

       @param arcsRecord ARCS쥳
    */
    private void addArcsRecord(final ArcsRecord arcsRecord)
	throws CorruptedFileException {
	int startIndex = arcsRecord.getStartIndex();
	ArcRecord[] list = arcsRecord.getList();
	if (startIndex >= blocks.length) {
	    throw new CorruptedFileException();
	}
	for (ArcRecord arcRecord : list) {
	    int endIndex = arcRecord.getEndIndex();
	    int flags = arcRecord.getFlags();
	    if (endIndex >= blocks.length) {
		throw new CorruptedFileException();
	    }
	    Block start = blocks[startIndex];
	    Block end = blocks[endIndex];
	    Arc arc = new Arc(start, end, flags);
	    if (!arc.isOnTree()) {
		/*
		  ѥ˥󥰥ĥ꡼ǤϤʤgcdaեˤϤ
		  б¹ԲϿ롣μ¹Բ
		  ºݤ˲褹ΤsetFunctionDataRecord()᥽
		  ɤƤФ줿Ȥ
		*/
		solvedArcs.add(arc);
	    } else {
		/*
		  ѥ˥󥰥ĥ꡼Υμ¹Բgcdaե
		  뤫ǤʤΤǡեդ򤯤Ȥǥ
		  μ¹ԲʤФʤʤ
		*/
		unsolvedArcs.add(arc);
	    }
	    if (arc.isFake()) {
		++fakeArcCount;
	    }
	}
	totalArcCount += list.length;
    }

    /**
       LINES쥳ɤԥȥΥꥹȤơؿդɲ
       ޤ

       @param linesRecord LINES쥳
    */
    private void addLinesRecord(final LinesRecord linesRecord)
	throws CorruptedFileException {
	int blockIndex = linesRecord.getBlockIndex();
	LineRecord[] list = linesRecord.getList();
	if (blockIndex >= blocks.length) {
	    throw new CorruptedFileException();
	}
	LineEntryList entryList = new LineEntryList(sourceFile);
	for (LineRecord rec : list) {
	    int number = rec.getNumber();
	    if (number == 0) {
		entryList.changeFileName(rec.getFileName());
	    } else {
		entryList.addLineNumber(number);
	    }
	}
	blocks[blockIndex].setLines(entryList.getLineEntries());
    }

    /**
       եդ򤭤ޤ

       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    private void solveFlowGraph() throws CorruptedFileException {
	LinkedList<Block> invalidBlocks = new LinkedList<Block>();
	for (Block e : blocks) {
	    e.sortOutArcs();
	    invalidBlocks.add(e);
	}

	LinkedList<Block> validBlocks = new LinkedList<Block>();
	Block e;
	while ((e = invalidBlocks.poll()) != null) {
	    e.validate(validBlocks);
	}
	while ((e = validBlocks.poll()) != null) {
	    e.validateSides(validBlocks);
	}
	for (Block b : blocks) {
	    if (!b.getCountValid()) {
		throw new CorruptedFileException("graph is unsolvable");
	    }
	}
    }

    /**
       ؿθƤӽФ¹Ԥ줿֥åȽи
       ˤ׻ޤ

       calledCount, returnedCount, executedBlockCountͭˤʤޤ
    */
    private void countCallSummary() {
	calledCount = blocks[0].getCount();

	ArrayList<Arc> list = blocks[blocks.length - 1].getInArcs();
	long count = 0;
	for (Arc arc : list) {
	    if (arc.isFake()) {
		continue;
	    }
	    count += arc.getCount();
	}
	returnedCount = count;

	int start = 1;
	int end = blocks.length - 1;
	for (int k = start; k < end; ++k) {
	    if (blocks[k].getCount() > 0) {
		++executedBlockCount;
	    }
	}
    }

    /**
       ؿǡ쥳ɤؿդɲäޤ

       @param rec ؿǡ쥳
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    public void setFunctionDataRecord(final FunctionDataRecord rec)
	throws CorruptedFileException {
	if (checksum != rec.getChecksum()) {
	    String m = String.format("gcda file: checksum mismatch for '%s'",
				     functionName);
	    throw new CorruptedFileException(m);
	}
	long[] arcCounts = rec.getArcCounts();
	if (solvedArcs.size() != arcCounts.length) {
	    String m = String.format("gcda file: profile mismatch for '%s'",
				     functionName);
	    throw new CorruptedFileException(m);
	}
	for (int k = 0; k < arcCounts.length; ++k) {
	    solvedArcs.get(k).addCount(arcCounts[k]);
	}
	solveFlowGraph();
	countCallSummary();
	solved = true;
    }

    /**
       ̻Ҥޤ

       @return ̻
    */
    public int getId() {
	return id;
    }

    /**
       ؿϤޤֹޤ

       @return ؿϤޤֹ
    */
    public int getLineNumber() {
	return lineNumber;
    }

    /**
       ؿ̾ޤ

       @return ؿ̾
    */
    public String getFunctionName() {
	return functionName;
    }

    /**
       ؿƤФ줿ޤ

       setFunctionDataRecord()Ǵؿǡ쥳ɤꤵƤ
       ɬפޤ

       @return ؿƤФ줿
    */
    public long getCalledCount() {
	return calledCount;
    }

    /**
       ؿäޤ

       setFunctionDataRecord()Ǵؿǡ쥳ɤꤵƤ
       ɬפޤ

       @return ؿä
    */
    public long getReturnedCount() {
	return returnedCount;
    }

    /**
       и¹Ԥ줿֥åޤ

       setFunctionDataRecord()Ǵؿǡ쥳ɤꤵƤ
       ɬפޤ

       @return ¹Ԥ줿֥åи
    */
    public int getExecutedBlockCount() {
	return executedBlockCount;
    }

    /**
       и֥åޤ

       @return ֥åи
    */
    public int getBlockCount() {
	return blocks.length - 2;
    }

    /**
       ؿϤޤֹӤ륳ѥ졼Ǥ
    */
    private static Comparator<FunctionGraph> lineNumberComparator;

    static {
	lineNumberComparator = new Comparator<FunctionGraph>() {
	    public int compare(final FunctionGraph fg1,
			       final FunctionGraph fg2) {
		return fg1.lineNumber - fg2.lineNumber;
	    }
	};
    }

    /**
       ؿϤޤֹӤ륳ѥ졼֤ޤ

       @return ؿϤޤֹӤ륳ѥ졼
    */
    public static Comparator<FunctionGraph> getLineNumberComparator() {
	return lineNumberComparator;
    }
}
