package com.maroontress.coverture;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.HashMap;
import java.util.TreeSet;

/**
   Хå륽եޤ
*/
public final class Source {

    /** ХåоݤΥեΥѥǤ */
    private String sourceFile;

    /** ֹȹԾΥޥåפǤ */
    private HashMap<Integer, LineInfo> map;

    /** ե˴ޤޤؿδؿդΥåȤǤ */
    private TreeSet<FunctionGraph> functions;

    /**
       ޤ

       @param sourceFile եΥѥ
    */
    public Source(final String sourceFile) {
	this.sourceFile = sourceFile;
	map = new HashMap<Integer, LineInfo>();
	functions = new TreeSet<FunctionGraph>(
	    FunctionGraph.getLineNumberComparator());
    }

    /**
       Υ˴ޤޤؿդɲäޤ

       @param fg ؿ
    */
    public void addFunctionGraph(final FunctionGraph fg) {
	functions.add(fg);
    }

    /**
       ¹ԲǽʹֹΤιֹμ¹Բûޤ

       @param lineNuber ֹ
       @param delta ¹Բ
    */
    public void addLineCount(final int lineNuber, final long delta) {
	LineInfo info = map.get(lineNuber);
	if (info == null) {
	    info = new LineInfo();
	    map.put(lineNuber, info);
	}
	info.addCount(delta);
    }

    /**
       ֹμ¹Բޤ

       ιֹ椬¹ԲǽǤʤ-1֤ޤ

       @param lineNuber ֹ
       @return ¹Բ
    */
    public long getLineCount(final int lineNuber) {
	LineInfo info = map.get(lineNuber);
	if (info == null) {
	    return -1;
	}
	return info.getCount();
    }

    /**
       ¹ԤԿޤ

       @return ¹ԤԿ
    */
    public int getExecutedLines() {
	Collection<LineInfo> all = map.values();
	int count = 0;
	for (LineInfo i : all) {
	    if (i.getCount() > 0) {
		++count;
	    }
	}
	return count;
    }

    /**
       ¹ԲǽʹԿޤ

       @return ¹ԲǽʹԿ
    */
    public int getExecutableLines() {
	return map.size();
    }

    /**
       եΥѥޤ

       @return եΥѥ
    */
    public String getPath() {
	return sourceFile;
    }

    /**
       ΨѡȤ׻ޤʬ줬0ΤȤ0֤ޤ

       @param n ʬ
       @param m ʬ
       @return n/mΥѡ
    */
    private int percent(final long n, final long m) {
	if (m == 0) {
	    return 0;
	}
	return (int) (100.0 * n / m + 0.5);
    }

    /**
       gcovߴΥХå̤Ϥޤ

       @param out 
       @param in եΥ꡼
       @throws IOException ϥ顼
    */
    private void outputGcovFile(final PrintWriter out,
				final LineNumberReader in) throws IOException {
	String line;
	Traverser<FunctionGraph> tr = new Traverser<FunctionGraph>(functions);
	while ((line = in.readLine()) != null) {
	    int num = in.getLineNumber();
	    while (tr.peek() != null && tr.peek().getLineNumber() == num) {
		FunctionGraph fg = tr.poll();
		long calledCount = fg.getCalledCount();
		long returnedCount = fg.getReturnedCount();
		int executedBlocks = fg.getExecutedBlockCount();
		int allBlocks = fg.getBlockCount();
		out.printf("function %s called %d returned %d%%"
			   + " blocks executed %d%%\n",
			   fg.getFunctionName(), calledCount,
			   percent(returnedCount, calledCount),
			   percent(executedBlocks, allBlocks));
	    }

	    LineInfo info = map.get(num);
	    long count;
	    String mark;
	    if (info == null) {
		mark = "-";
	    } else if ((count = info.getCount()) == 0) {
		mark = "#####";
	    } else {
		mark = String.valueOf(count);
	    }
	    out.printf("%9s:%5d:%s\n", mark, num, line);
	}
    }

    /**
       gcovߴΥХå̤Ϥޤ

       @param out 
       @param origin gcnoեΥꥸ
       @param prop ϥץѥƥ
       @throws IOException ϥ顼
    */
    private void outputLines(final PrintWriter out, final Origin origin,
			     final IOProperties prop) throws IOException {
	File file = new File(sourceFile);
	if (file.lastModified() > origin.getNoteFile().lastModified()) {
	    System.err.printf("%s: source file is newer than gcno file\n",
			      sourceFile);
	    out.printf("%9s:%5d:Source is newer than gcno file\n", "-", 0);
	}
	LineNumberReader in
	    = new LineNumberReader(prop.createSourceFileReader(file));
	try {
	    outputGcovFile(out, in);
	} finally {
	    in.close();
	}
    }

    /**
       Хåեޤ

       @param origin gcnoեΥꥸ
       @param runs ץμ¹Բ
       @param programs ץο
       @param prop ϥץѥƥ
       @throws IOException ϥ顼
    */
    public void outputFile(final Origin origin, final int runs,
			   final int programs, final IOProperties prop)
	throws IOException {
	String path = origin.getCoverageFilePath(sourceFile);
	PrintWriter out;
	try {
	    out = new PrintWriter(prop.createGcovWriter(path));
	} catch (FileNotFoundException e) {
	    File gcov = prop.createOutputFile(path);
	    System.err.printf("%s: can't open: %s\n",
			      gcov.getPath(), e.getMessage());
	    return;
	}
	try {
	    File gcnoFile = origin.getNoteFile();
	    File gcdaFile = origin.getDataFile();
	    out.printf("%9s:%5d:Source:%s\n", "-", 0, sourceFile);
	    out.printf("%9s:%5d:Graph:%s\n", "-", 0, gcnoFile.getPath());
	    out.printf("%9s:%5d:Data:%s\n", "-", 0, gcdaFile.getPath());
	    out.printf("%9s:%5d:Runs:%d\n", "-", 0, runs);
	    out.printf("%9s:%5d:Programs:%d\n", "-", 0, programs);
	    outputLines(out, origin, prop);
	} finally {
	    out.close();
	}
	if (prop.isVerbose()) {
	    File gcov = prop.createOutputFile(path);
	    System.err.printf("%s: created.\n", gcov.getPath());
	}
    }
}
